@@ -33,6 +33,7 @@
 #include "eel-gtk-macros.h"
 #include "eel-lib-self-check-functions.h"
 #include "eel-string.h"
+#include <gtk/gtkprivate.h>
 #include <gtk/gtkselection.h>
 #include <gtk/gtksignal.h>
 #include <libart_lgpl/art_rgb.h>
@@ -104,6 +105,7 @@ struct EelBackgroundDetails {
 	int background_entire_height;
 	GdkColor background_color;
 	gboolean background_changes_with_size;
+	guint respond_to_style_set : 1;
 };
 
 static void
@@ -169,6 +171,8 @@ eel_background_init (gpointer object, gpointer klass)
 
 	background->details = g_new0 (EelBackgroundDetails, 1);
 	background->details->constant_size = FALSE;
+	background->details->is_solid_color = TRUE;
+	background->details->respond_to_style_set = TRUE;
 }
 
 /* The safe way to clear an image from a background is:
@@ -634,13 +638,29 @@ static void
 eel_background_ensure_realized (EelBackground *background, GdkWindow *window,
 				int entire_width, int entire_height)
 {
+	GdkColor color;
 	int pixmap_width, pixmap_height;
 	char *start_color_spec;
  	GdkPixmap *pixmap;
 	GdkGC *gc;
+	GtkWidget *widget;
+	GtkStyle *style;
+
+	/* Try to parse the color spec.  If we fail, default to the style's color */
 
 	start_color_spec = eel_gradient_get_start_color_spec (background->details->color);
-	eel_gdk_color_parse_with_white_default (start_color_spec, &background->details->background_color);
+
+	if (start_color_spec && eel_gdk_color_parse (start_color_spec, &color))
+		background->details->background_color = color;
+	else {
+		/* Get the widget to which the window belongs and its style as well */
+		gdk_window_get_user_data (window, (void **) &widget);
+		g_assert (widget != NULL);
+
+		style = gtk_widget_get_style (widget);
+		background->details->background_color = style->bg[GTK_STATE_NORMAL];
+	}
+
 	g_free (start_color_spec);
 
 	/* If the pixmap doesn't change with the window size, never update
@@ -985,7 +1005,9 @@ eel_background_start_loading_image (EelBackground *background, gboolean emit_app
 			eel_background_load_image_callback (0, pixbuf, background);
 
 			/* load_image_callback refs the pixbuf so we unref it here */
-			g_object_unref (pixbuf);
+			if (pixbuf != NULL) {
+				g_object_unref (pixbuf);
+			}
 		}
 	} else {
 		eel_background_load_image_callback (0, NULL, background);
@@ -1103,18 +1125,56 @@ eel_background_receive_dropped_background_image (EelBackground *background,
 	eel_background_set_image_uri_and_color (background, image_uri, NULL);
 }
 
+static void
+eel_background_set_style_font_from_default (GtkStyle *style,
+					    GtkSettings *settings)
+{
+	char *fontname;
+
+	fontname = NULL;
+	
+	g_object_get (settings,
+		      "gtk-font-name", &fontname,
+		      NULL);
+
+	if (fontname != NULL) {
+		if (style->font_desc) {
+			pango_font_description_free (style->font_desc);
+		}
+		
+		style->font_desc = pango_font_description_from_string (fontname);
+		g_free (fontname);
+	}
+}
+
+
 static void
 eel_background_set_widget_style (EelBackground *background,
 				 GtkWidget *widget)
 {
 	GtkStyle *style;
+	gboolean old_respond_value;
 	
 	g_return_if_fail (EEL_IS_BACKGROUND (background));
 	g_return_if_fail (GTK_IS_WIDGET (widget));
 	
 	style = GTK_STYLE (eel_background_style_new (gtk_widget_get_style (widget),
-						     background));	
+						     background));
+	
+	/* Bad hack to make font changing work: */
+	eel_background_set_style_font_from_default (style, gtk_widget_get_settings (widget));
+
+	old_respond_value = background->details->respond_to_style_set;
+	background->details->respond_to_style_set = FALSE;
 	gtk_widget_set_style (widget, style);
+	/* HACK: unset the USER_STYLE flag because otherwise we will not be
+	 * notified of theme changes through the "style_set" signal.  The real
+	 * solution is not to use a style of our own, but rather paint the
+	 * widget by hand when needed.
+	 */
+	GTK_PRIVATE_UNSET_FLAG (widget, GTK_USER_STYLE);
+	background->details->respond_to_style_set = old_respond_value;
+
 	g_object_unref (style);
 }
 
@@ -1255,6 +1315,52 @@ eel_widget_background_changed (GtkWidget *widget, EelBackground *background)
 	gtk_widget_queue_draw (widget);
 }
 
+/* Bad hack to make font changes work for EelBackgrounds */
+static void
+eel_widget_font_changed (GtkWidget *widget,
+			 GParamSpec *pspec,
+			 GtkSettings *settings)
+{
+	EelBackground *background;
+	GtkStyle *style;
+	gboolean old_respond_value;
+
+	background = eel_get_widget_background (widget);
+
+	style = gtk_style_copy (gtk_widget_get_style (widget));
+	
+	eel_background_set_style_font_from_default (style, settings);
+		
+	old_respond_value = background->details->respond_to_style_set;
+	background->details->respond_to_style_set = FALSE;
+	gtk_widget_set_style (widget, style);
+	/* HACK: unset the USER_STYLE flag because otherwise we will not be
+	 * notified of theme changes through the "style_set" signal.  The real
+	 * solution is not to use a style of our own, but rather paint the
+	 * widget by hand when needed.
+	 */
+	GTK_PRIVATE_UNSET_FLAG (widget, GTK_USER_STYLE);
+	background->details->respond_to_style_set = old_respond_value;
+
+	g_object_unref (style);
+}
+
+/* Callback used when the style of a widget changes.  We have to regenerate its
+ * EelBackgroundStyle so that it will match the chosen GTK+ theme.
+ */
+static void
+widget_style_set_cb (GtkWidget *widget, GtkStyle *previous_style, gpointer data)
+{
+	EelBackground *background;
+
+	background = EEL_BACKGROUND (data);
+
+	if (!background->details->respond_to_style_set)
+		return;
+
+	eel_widget_background_changed (widget, background);
+}
+
 /* Gets the background attached to a widget.
 
    If the widget doesn't already have a EelBackground object,
@@ -1270,6 +1376,7 @@ eel_get_widget_background (GtkWidget *widget)
 {
 	gpointer data;
 	EelBackground *background;
+	GtkSettings *settings;
 
 	g_return_val_if_fail (GTK_IS_WIDGET (widget), NULL);
 
@@ -1292,6 +1399,14 @@ eel_get_widget_background (GtkWidget *widget)
 				 G_CALLBACK (eel_widget_background_changed), widget, G_CONNECT_SWAPPED);
 	eel_widget_background_changed (widget, background);
 
+	settings = gtk_widget_get_settings (widget);
+	g_signal_connect_object (settings, "notify::gtk-font-name",
+				 G_CALLBACK (eel_widget_font_changed), widget, G_CONNECT_SWAPPED);
+
+	g_signal_connect (widget, "style_set",
+			  G_CALLBACK (widget_style_set_cb),
+			  background);
+
 	return background;
 }
 
